use crate::{
    payment::PaymentTo,
    storage::{self, Error, Status},
    transaction::inner::TransactionInner,
    AccountId, PaymentId, Transaction, TransactionId, Type,
};
use sqlx::{Row, Sqlite, Transaction as SqlxTransaction};
use std::{marker::PhantomData, num::TryFromIntError};

/// Creates a new Batch for SQLite
///
/// Batches are a group of updates to the databases, in which all of the are
/// executed or none. Same concept as a rdbms transaction.
pub struct Batch<'a> {
    inner: SqlxTransaction<'a, Sqlite>,
    x: PhantomData<&'a ()>,
}

impl<'a> Batch<'a> {
    /// Creates a new instance
    pub fn new(inner: SqlxTransaction<'a, Sqlite>) -> Batch<'a> {
        Self {
            inner,
            x: PhantomData,
        }
    }
}

#[async_trait::async_trait]
impl<'a> storage::Batch<'a> for Batch<'a> {
    async fn rollback(self) -> Result<(), Error> {
        self.inner
            .rollback()
            .await
            .map_err(|e| Error::Storage(e.to_string()))
    }

    async fn relate_account_to_transaction(
        &mut self,
        transaction: &TransactionId,
        account: &AccountId,
        typ: Type,
    ) -> Result<(), Error> {
        sqlx::query(
            r#"
            INSERT INTO "transaction_accounts" ("transaction_id", "account_id", "type")
            VALUES (?, ?, ?)
            "#,
        )
        .bind(transaction.to_string())
        .bind(account.to_string())
        .bind::<u32>(typ.into())
        .execute(&mut *self.inner)
        .await
        .map_err(|e| Error::Storage(e.to_string()))?;
        Ok(())
    }

    async fn get_transaction(
        &mut self,
        transaction_id: &TransactionId,
    ) -> Result<Transaction, Error> {
        let row = sqlx::query(
            r#"
            SELECT
                "b"."blob"
            FROM
                "transactions" as "t",
                "blobs" as "b"
            WHERE
                "t"."transaction_id" = ?
                AND "t"."blob_id" = "b"."id"
        "#,
        )
        .bind(transaction_id.to_string())
        .fetch_one(&mut *self.inner)
        .await
        .map_err(|e| Error::Storage(e.to_string()))?;

        let encoded = row
            .try_get::<Vec<u8>, usize>(0)
            .map_err(|e| Error::Storage(e.to_string()))?;
        bincode::deserialize::<TransactionInner>(&encoded)
            .map_err(|e| Error::Storage(e.to_string()))?
            .try_into()
            .map_err(|e: crate::transaction::Error| Error::Storage(e.to_string()))
    }

    async fn commit(self) -> Result<(), Error> {
        self.inner
            .commit()
            .await
            .map_err(|e| Error::Storage(e.to_string()))
    }

    async fn spend_payments(
        &mut self,
        transaction_id: &TransactionId,
        spends: Vec<PaymentId>,
        status: Status,
    ) -> Result<(), Error> {
        let spend_ids = spends.iter().map(|id| id.to_string()).collect::<Vec<_>>();
        let placeholder = format!("?{}", ", ?".repeat(spend_ids.len()));

        let sql = format!(
            r#"
            UPDATE "payments"
            SET "status" = ?, "spent_by" = ?
            WHERE "payment_id" IN ({placeholder})
            AND "spent_by" IS NULL
        "#
        );

        let mut query = sqlx::query(&sql)
            .bind::<u32>(status.into())
            .bind(transaction_id.to_string());

        for id in spend_ids {
            query = query.bind(id);
        }

        let updated_records = usize::try_from(
            query
                .execute(&mut *self.inner)
                .await
                .map_err(|e| Error::Storage(e.to_string()))?
                .rows_affected(),
        )
        .map_err(|_| Error::Overflow)?;

        if updated_records == spends.len() {
            Ok(())
        } else {
            Err(Error::NoUpdate)
        }
    }

    async fn create_payments(
        &mut self,
        transaction_id: &TransactionId,
        recipients: &[PaymentTo],
        status: Status,
    ) -> Result<(), Error> {
        for (pos, recipient) in recipients.iter().enumerate() {
            sqlx::query(
                r#"
                INSERT INTO "payments"("payment_id", "to", "status", "asset", "cents")
                VALUES(?, ?, ?, ?, ? )
            "#,
            )
            .bind(
                PaymentId {
                    transaction: transaction_id.clone(),
                    position: pos
                        .try_into()
                        .map_err(|e: TryFromIntError| Error::Encoding(e.to_string()))?,
                }
                .to_string(),
            )
            .bind(recipient.to.to_string())
            .bind::<u32>(status.into())
            .bind(recipient.amount.asset().to_string())
            .bind(recipient.amount.cents().to_string())
            .execute(&mut *self.inner)
            .await
            .map_err(|e| Error::Storage(e.to_string()))?;
        }

        Ok(())
    }

    async fn update_transaction_payments(
        &mut self,
        transaction_id: &TransactionId,
        create_status: Status,
        spend_status: Status,
    ) -> Result<(usize, usize), Error> {
        let creates = sqlx::query(
            r#"
            UPDATE "payments"
            SET "status" = ?, updated_at = datetime('now')
            WHERE "payment_id" LIKE ? AND "status" IN (?, ?)
        "#,
        )
        .bind::<u32>(create_status.into())
        .bind(format!("{}:%", transaction_id))
        .bind::<u32>(Status::Spendable.into())
        .bind::<u32>(Status::Locked.into())
        .execute(&mut *self.inner)
        .await
        .map_err(|e| Error::Storage(e.to_string()))?;

        let spends = sqlx::query(
            r#"
            UPDATE "payments"
            SET "status" = ?, "spent_by" = ?, updated_at = datetime('now')
            WHERE "spent_by" = ? AND "status" IN (?, ?)
            "#,
        )
        .bind::<u32>(spend_status.into())
        .bind(if spend_status == Status::Spendable {
            None
        } else {
            Some(transaction_id.to_string())
        })
        .bind(transaction_id.to_string())
        .bind::<u32>(Status::Spendable.into())
        .bind::<u32>(Status::Locked.into())
        .execute(&mut *self.inner)
        .await
        .map_err(|e| Error::Storage(e.to_string()))?;

        Ok((
            usize::try_from(creates.rows_affected()).map_err(|_| Error::Overflow)?,
            usize::try_from(spends.rows_affected()).map_err(|_| Error::Overflow)?,
        ))
    }

    async fn get_payment_status(
        &mut self,
        transaction_id: &TransactionId,
    ) -> Result<Option<Status>, Error> {
        let row = sqlx::query(
            r#"
            SELECT
                "p"."status"
            FROM
                "payments" "p"
            WHERE
                "p"."payment_id" = ?
            LIMIT 1
            "#,
        )
        .bind(transaction_id.to_string())
        .fetch_optional(&mut *self.inner)
        .await
        .map_err(|e| Error::Storage(e.to_string()))?;

        if let Some(row) = row {
            let status = row
                .try_get::<u32, usize>(0)
                .map_err(|_| Error::Storage("failed to parse status".to_owned()))?;

            status
                .try_into()
                .map(Some)
                .map_err(|_| Error::Storage("failed to parse status".to_owned()))
        } else {
            return Ok(None);
        }
    }

    async fn store_transaction(&mut self, transaction: &Transaction) -> Result<(), Error> {
        sqlx::query(
            r#"
                INSERT INTO "blobs"("id", "transaction_id", "blob", "status", "created_at")
                VALUES(?, ?, ?, ?, ?)
            "#,
        )
        .bind(transaction.revision.to_string())
        .bind(transaction.id.to_string())
        .bind(bincode::serialize(transaction.inner()).map_err(|e| Error::Encoding(e.to_string()))?)
        .bind(transaction.status().to_string())
        .bind(transaction.created_at())
        .execute(&mut *self.inner)
        .await
        .map_err(|e| Error::Storage(e.to_string()))?;

        sqlx::query(
            r#"
                INSERT INTO "transactions"("transaction_id", "blob_id", "status")
                VALUES(?, ?, ?)
                ON CONFLICT("transaction_id")
                    DO UPDATE SET "blob_id" = EXCLUDED."blob_id", "status" = EXCLUDED."status"
            "#,
        )
        .bind(transaction.id.to_string())
        .bind(transaction.revision.to_string())
        .bind(transaction.status().to_string())
        .execute(&mut *self.inner)
        .await
        .map_err(|e| Error::Storage(e.to_string()))?;

        Ok(())
    }

    async fn tag_transaction(
        &mut self,
        _transaction: &Transaction,
        _tags: &[String],
    ) -> Result<(), Error> {
        todo!()
    }
}
