//! SQLite storage layer for Verax
use crate::{
    amount::AmountCents,
    storage::{Error, Storage},
    transaction::{inner::TransactionInner, Type},
    AccountId, Amount, Asset, PaymentFrom, PaymentId, Transaction, TransactionId,
};
use futures::TryStreamExt;
use sqlx::{sqlite::SqliteRow, Executor, Row};
use std::collections::HashMap;

mod batch;

pub use batch::Batch;
pub use sqlx::sqlite::{SqliteConnectOptions, SqlitePoolOptions};

use super::Status;

/// SQLite storage layer for Verax
pub struct SQLite {
    db: sqlx::SqlitePool,
}

impl SQLite {
    /// Creates a new Verax SQLite storage layer. It takes an instance of the asset_manager
    pub fn new(db: sqlx::SqlitePool) -> Self {
        Self { db }
    }

    /// Creates all the tables and indexes that are needed
    /// The database uses string for the primary keys to make things easier to debug[1]
    ///
    /// [1] https://sqlite.org/forum/info/69c8cbbe2d84410f
    pub async fn setup(&self) -> Result<(), sqlx::Error> {
        let mut x = self.db.begin().await?;
        x.execute(
            r#"
        CREATE TABLE IF NOT EXISTS "transactions" (
            "transaction_id" VARCHAR(66) NOT NULL PRIMARY KEY,
            "blob_id" VARCHAR(66) NOT NULL,
            "status" VARCHAR(10) NOT NULL,
            "created_at" DATETIME DEFAULT CURRENT_TIMESTAMP,
            "updated_at" DATETIME DEFAULT CURRENT_TIMESTAMP
        );
        CREATE TABLE IF NOT EXISTS "blobs" (
            "id" VARCHAR(64) NOT NULL PRIMARY KEY,
            "transaction_id" VARCHAR(66) NOT NULL,
            "previous_blob_id" BLOB,
            "status" VARCHAR(30) NOT NULL,
            "blob" BINARY NOT NULL,
            "created_at" DATETIME DEFAULT CURRENT_TIMESTAMP
        );
        CREATE TABLE IF NOT EXISTS "payments" (
            "payment_id" VARCHAR(70) NOT NULL PRIMARY KEY,
            "to" VARCHAR(255) NOT NULL,
            "status" INT NOT NULL,
            "asset" VARCHAR(10) NOT NULL,
            "cents" STRING NOT NULL,
            "spent_by" VARCHAR(64) DEFAULT NULL,
            "created_at" DATETIME DEFAULT CURRENT_TIMESTAMP,
            "updated_at" DATETIME DEFAULT CURRENT_TIMESTAMP
        );
        CREATE INDEX IF NOT EXISTS "spent_by" ON "payments" ("to", "spent_by");
        CREATE TABLE IF NOT EXISTS "transaction_accounts" (
            "account_id" VARCHAR(64) NOT NULL,
            "transaction_id" VARCHAR(66) NOT NULL,
            "type" INTEGE
            R NOT NULL,
            "created_at" DATETIME DEFAULT CURRENT_TIMESTAMP,
            "updated_at" DATETIME DEFAULT CURRENT_TIMESTAMP,
            PRIMARY KEY("account_id", "transaction_id")
        );
        "#,
        )
        .await
        .expect("valid");
        x.commit().await?;
        Ok(())
    }

    fn sql_row_to_amount(row: &SqliteRow, pos: usize) -> Result<Amount, Error> {
        let asset: Asset = row
            .try_get::<String, usize>(pos)
            .map_err(|_| Error::Storage("Invalid asset".to_string()))?
            .parse()
            .map_err(|_| Error::Storage("Invalid asset".to_string()))?;

        let cents = row
            .try_get::<i64, usize>(pos.checked_add(1).ok_or(Error::Overflow)?)
            .map_err(|_| Error::Storage("Invalid cents".to_string()))?
            .into();

        Ok(asset.new_amount(cents))
    }

    fn sql_row_to_payment_id(row: &SqliteRow, pos: usize) -> Result<PaymentId, Error> {
        row.try_get::<String, usize>(pos)
            .map_err(|_| Error::Storage("Invalid payment_id".to_string()))?
            .as_str()
            .parse()
            .map_err(|_| Error::Storage("Invalid payment_id length".to_string()))
    }
}

#[async_trait::async_trait]
impl Storage for SQLite {
    type Batch<'a> = Batch<'a>
    where
        Self: 'a;

    async fn begin<'a>(&'a self) -> Result<Self::Batch<'a>, Error> {
        self.db
            .begin()
            .await
            .map(Batch::new)
            .map_err(|x| Error::Storage(x.to_string()))
    }

    async fn get_balance(&self, account: &AccountId) -> Result<Vec<Amount>, Error> {
        let mut conn = self
            .db
            .acquire()
            .await
            .map_err(|e| Error::Storage(e.to_string()))?;

        let mut result = sqlx::query(
            r#"
            SELECT
                "asset",
                SUM("cents") as "cents"
            FROM
                "payments"
            WHERE
                "to" = ? AND "spent_by" IS NULL AND status = ?
            GROUP BY "asset"
            "#,
        )
        .bind(account.to_string())
        .bind::<u32>(Status::Spendable.into())
        .fetch(&mut *conn);

        let mut balances = HashMap::<Asset, Amount>::new();

        while let Some(row) = result
            .try_next()
            .await
            .map_err(|e| Error::Storage(e.to_string()))?
        {
            let asset: Asset = row
                .try_get::<String, usize>(0)
                .map_err(|_| Error::Storage("Invalid asset_id".to_string()))?
                .parse()
                .map_err(|_| Error::Storage("Invalid asset_id".to_string()))?;

            let cents = row
                .try_get::<i64, usize>(1)
                .map_err(|_| Error::Storage("Invalid cents".to_string()))?
                .into();

            let new_amount = asset.new_amount(cents);

            if let Some(amount) = balances.get_mut(&asset) {
                *amount = amount
                    .checked_add(&new_amount)
                    .ok_or(Error::Storage("amount overflow".to_owned()))?;
            } else {
                balances.insert(asset, new_amount);
            }
        }

        Ok(balances.into_values().collect())
    }

    async fn get_unspent_payments(
        &self,
        account: &AccountId,
        asset: &Asset,
        mut target_amount: AmountCents,
    ) -> Result<Vec<PaymentFrom>, Error> {
        let mut conn = self
            .db
            .acquire()
            .await
            .map_err(|e| Error::Storage(e.to_string()))?;
        let results = sqlx::query(
            r#"
            SELECT
                "p"."payment_id",
                "p"."asset",
                "p"."cents",
                "p"."to"
            FROM
                "payments" as "p"
            WHERE
                "p"."to" = ? AND "p"."asset" = ? AND status = ? AND "p"."spent_by" IS NULL
            ORDER BY cents ASC
            "#,
        )
        .bind(account.to_string())
        .bind(asset.to_string())
        .bind::<u32>(Status::Spendable.into())
        .fetch_all(&mut *conn)
        .await
        .map_err(|e| Error::Storage(e.to_string()))?;

        let mut to_return = vec![];
        for row in results.into_iter() {
            let amount = Self::sql_row_to_amount(&row, 1)?;
            target_amount = target_amount
                .checked_sub(amount.cents())
                .ok_or(Error::Underflow)?;
            to_return.push(PaymentFrom {
                id: Self::sql_row_to_payment_id(&row, 0)?,
                from: row
                    .try_get::<String, usize>(3)
                    .map_err(|_| Error::Storage("Invalid from".to_string()))?
                    .parse()
                    .map_err(|_| Error::Encoding("Invalid account encoding".to_owned()))?,
                amount,
            });
            if target_amount <= 0 {
                break;
            }
        }

        if target_amount <= 0 {
            Ok(to_return)
        } else {
            Err(Error::NotEnoughUnspentPayments(target_amount))
        }
    }

    async fn get_transaction(&self, transaction_id: &TransactionId) -> Result<Transaction, Error> {
        let mut conn = self
            .db
            .acquire()
            .await
            .map_err(|e| Error::Storage(e.to_string()))?;

        let row = sqlx::query(
            r#"
            SELECT
                "b"."blob"
            FROM
                "transactions" as "t",
                "blobs" as "b"
            WHERE
                "t"."transaction_id" = ?
                AND "t"."blob_id" = "b"."id"
        "#,
        )
        .bind(transaction_id.to_string())
        .fetch_one(&mut *conn)
        .await
        .map_err(|e| Error::Storage(e.to_string()))?;

        let encoded = row
            .try_get::<Vec<u8>, usize>(0)
            .map_err(|e| Error::Storage(e.to_string()))?;

        bincode::deserialize::<TransactionInner>(&encoded)
            .map_err(|e| Error::Storage(e.to_string()))?
            .try_into()
            .map_err(|e: crate::transaction::Error| Error::Storage(e.to_string()))
    }

    async fn get_transactions(
        &self,
        account: &AccountId,
        types: &[Type],
        _tags: &[String],
    ) -> Result<Vec<Transaction>, Error> {
        let mut conn = self
            .db
            .acquire()
            .await
            .map_err(|e| Error::Storage(e.to_string()))?;

        let sql = if types.is_empty() {
            r#"SELECT "transaction_id" FROM "transaction_accounts" WHERE "account_id" = ? ORDER BY "created_at" DESC"#.to_owned()
        } else {
            let params = format!(
                "?{}",
                ", ?".repeat(types.len().checked_add(1).ok_or(Error::Underflow)?)
            );
            format!(
                r#"SELECT "transaction_id" FROM "transaction_accounts" WHERE "account_id" = ? AND "type" IN ({}) ORDER BY "created_at" DESC"#,
                params
            )
        };

        let sql = sqlx::query(&sql).bind(account.to_string());

        let ids = if !types.is_empty() {
            let mut sql = sql;
            for typ in types.iter() {
                sql = sql.bind::<u32>(typ.into());
            }
            sql
        } else {
            sql
        }
        .fetch_all(&mut *conn)
        .await
        .map_err(|e| Error::Storage(e.to_string()))?
        .into_iter()
        .map(|row| {
            let id: Result<TransactionId, _> = row
                .try_get::<String, usize>(0)
                .map_err(|_| Error::Storage("Invalid transaction_id".to_string()))?
                .as_str()
                .try_into();

            id.map_err(|_| Error::Storage("Invalid transaction_id length".to_string()))
        })
        .collect::<Result<Vec<_>, Error>>()?;

        drop(conn);

        let mut transactions = vec![];
        for id in ids.into_iter() {
            transactions.push(self.get_transaction(&id).await?);
        }

        Ok(transactions)
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::storage_test_suite;
    use sqlx::sqlite::{SqliteConnectOptions, SqlitePoolOptions};
    use std::fs::remove_file;

    async fn get_instance(test: &str) -> SQLite {
        let path = format!("/tmp/{}.db", test);
        let _ = remove_file(&path);
        let settings = path
            .parse::<SqliteConnectOptions>()
            .expect("valid settings")
            .create_if_missing(true);

        let pool = SqlitePoolOptions::new()
            .connect_with(settings)
            .await
            .expect("pool");

        let db = SQLite::new(pool);
        db.setup().await.expect("setup");
        db
    }

    storage_test_suite!();
}
